<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Filesystem\File;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;
use Org\Mxchange\CoreFramework\EntryPoint\ApplicationEntryPoint;
use Org\Mxchange\CoreFramework\Factory\Object\ObjectFactory;
use Org\Mxchange\CoreFramework\Filesystem\File\BaseAbstractFile;
use Org\Mxchange\CoreFramework\Index\Indexable;
use Org\Mxchange\CoreFramework\Stack\File\StackableFile;
use Org\Mxchange\CoreFramework\Traits\Index\IndexableTrait;
use Org\Mxchange\CoreFramework\Traits\Stack\StackableTrait;

// Import SPL stuff
use \BadMethodCallException;
use \InvalidArgumentException;
use \LogicException;
use \OutOfBoundsException;
use \SplFileInfo;
use \UnexpectedValueException;

/**
 * A general binary file class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2020 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
abstract class BaseBinaryFile extends BaseAbstractFile implements BinaryFile {
	// Load traits
	use StackableTrait;
	use IndexableTrait;

	/**
	 * Separator for header data
	 */
	const SEPARATOR_HEADER_DATA = 0x01;

	/**
	 * Separator header->entries
	 */
	const SEPARATOR_HEADER_ENTRIES = 0x02;

	/**
	 * Separator group->hash
	 */
	const SEPARATOR_GROUP_HASH = 0x03;

	/**
	 * Separator hash->value
	 */
	const SEPARATOR_HASH_VALUE = 0x04;

	/**
	 * Separator entry->entry
	 */
	const SEPARATOR_ENTRIES = 0x05;

	/**
	 * Separator type->position
	 */
	const SEPARATOR_TYPE_POSITION = 0x06;

	/**
	 * Length of count
	 */
	const LENGTH_COUNT = 20;

	/**
	 * Length of position
	 */
	const LENGTH_POSITION = 20;

	/**
	 * Length of group
	 */
	const LENGTH_GROUP = 10;

	/**
	 * Maximum length of entry type
	 */
	const LENGTH_TYPE = 20;

	//***** Array elements for 'gaps' array *****

	/**
	 * Start of gap
	 */
	const GAPS_INDEX_START = 'start';

	/**
	 * End of gap
	 */
	const GAPS_INDEX_END = 'end';

	/**
	 * Current seek position
	 */
	private $seekPosition = 0;

	/**
	 * Size of header
	 */
	private $headerSize = 0;

	/**
	 * File header
	 */
	private $header = [];

	/**
	 * Seek positions for gaps ("fragmentation")
	 */
	private $gaps = [];

	/**
	 * Seek positions for damaged entries (e.g. mismatching hash sum, ...)
	 */
	private $damagedEntries = [];

	/**
	 * Back-buffer
	 */
	private $backBuffer = '';

	/**
	 * Currently loaded block (will be returned by current())
	 */
	private $currentBlock = '';

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the class
	 * @return	void
	 */
	protected function __construct (string $className) {
		// Call parent constructor
		parent::__construct($className);

		// Init counters and gaps array
		$this->initCountersGapsArray();
	}

	/**
	 * Setter for backBuffer field
	 *
	 * @param	$backBuffer		Characters to "store" in back-buffer
	 * @return	void
	 */
	private function setBackBuffer (string $backBuffer) {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Setting backBuffer(%d)=%s - CALLED!', strlen($backBuffer), $backBuffer));
		$this->backBuffer = $backBuffer;
	}

	/**
	 * Getter for backBuffer field
	 *
	 * @return	$backBuffer		Characters "stored" in back-buffer
	 */
	private function getBackBuffer () {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Getting this->backBuffer(%d)=%s - CALLED!', strlen($this->backBuffer), $this->backBuffer));
		return $this->backBuffer;
	}

	/**
	 * Setter for current field
	 *
	 * @param	$current	Characters to set a currently loaded block
	 * @return	void
	 */
	private function setCurrentBlock (string $currentBlock) {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Setting currentBlock(%d)=%s - CALLED!', strlen($currentBlock), $currentBlock));
		$this->currentBlock = $currentBlock;
	}

	/**
	 * Gets currently read data
	 *
	 * @return	$current	Currently read data
	 */
	public function getCurrentBlock () {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Getting this->currentBlock(%d)=%s - CALLED!', strlen($this->currentBlock), $this->currentBlock));
		return $this->currentBlock;
	}

	/**
	 * Getter for header size
	 *
	 * @return	$totalEntries	Size of file header
	 */
	public final function getHeaderSize () {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Getting this->headerSize=%d - CALLED!', $this->headerSize));
		return $this->headerSize;
	}

	/**
	 * Setter for header size
	 *
	 * @param	$headerSize		Size of file header
	 * @return	void
	 */
	public final function setHeaderSize (int $headerSize) {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Setting headerSize=%d - CALLED!', $headerSize));
		$this->headerSize = $headerSize;
	}

	/**
	 * Getter for header array
	 *
	 * @return	$totalEntries	Size of file header
	 */
	public final function getHeader () {
		// Get it
		return $this->header;
	}

	/**
	 * Setter for header
	 *
	 * @param	$header		Array for a file header
	 * @return	void
	 */
	public final function setHeader (array $header) {
		// Set it
		$this->header = $header;
	}

	/**
	 * Getter for seek position
	 *
	 * @return	$seekPosition	Current seek position (stored here in object)
	 */
	public final function getSeekPosition () {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Getting this->seekPosition=%d - CALLED!', $this->seekPosition));
		return $this->seekPosition;
	}

	/**
	 * Setter for seek position
	 *
	 * @param	$seekPosition	Current seek position (stored here in object)
	 * @return	void
	 */
	protected final function setSeekPosition (int $seekPosition) {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Setting seekPosition=%d - CALLED!', $seekPosition));
		$this->seekPosition = $seekPosition;
	}

	/**
	 * Marks whole file as gaps-only (freshly created file
	 *
	 * @param	$type	Type of file
	 * @param	$minimumBlockLength		Minimum block length
	 * @return	void
	 */
	private function markFileGapsOnly (string $type, int $minimumBlockLength) {
		// Very simple to do ...
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: type=%s,minimumBlockLength=%d - CALLED!', $type, $minimumBlockLength));
		for ($idx = 0; $idx < FrameworkBootstrap::getConfigurationInstance()->getConfigEntry($type . '_pre_allocate_count'); $idx++) {
			// Mark start and end position as gap
			$this->addGap($idx * $minimumBlockLength, $idx * $minimumBlockLength + $minimumBlockLength);
		}

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Checks whether the abstracted file only contains gaps by counting all
	 * gaps' bytes together and compare it to total length.
	 *
	 * @return	$isGapsOnly		Whether the abstracted file only contains gaps
	 */
	private function isFileGapsOnly () {
		// Count every gap
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$gapsSize = 0;
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: this->gaps()=%d', count($this->gaps)));
		foreach ($this->gaps as $gap) {
			// Calculate size of found gap: end-start including both
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: gap[%s]=%d,ga[%s]=%d', self::GAPS_INDEX_START, $gap[self::GAPS_INDEX_START], self::GAPS_INDEX_END, $gap[self::GAPS_INDEX_END]));
			$gapsSize += ($gap[self::GAPS_INDEX_END] - $gap[self::GAPS_INDEX_START]);

			// Debug message
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: gapsSize=%d', $gapsSize));
		}

		// Total gap size + header size must be same as file size
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: gapsSize=%d,this->fileSize=%d', $gapsSize, $this->getFileSize()));
		$isGapsOnly = ($gapsSize + 1 == $this->getFileSize());

		// Return status
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: isGapsOnly=%d - EXIT!', intval($isGapsOnly)));
		return $isGapsOnly;
	}

	/**
	 * Adds a gap for given start and end position
	 *
	 * @param	$startPosition	Start seek position
	 * @param	$endPosition	End seek position
	 * @return	void
	 */
	private function addGap(int $startPosition, int $endPosition) {
		// Push to gaps array
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: startPosition=%d,endPosition=%d - CALLED!', $startPosition, $endPosition));
		array_push($this->gaps, [
			self::GAPS_INDEX_START  => $startPosition,
			self::GAPS_INDEX_END    => $endPosition,
		]);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Initializes the back-buffer by setting it to an empty string.
	 *
	 * @return	void
	 */
	private function initBackBuffer () {
		// Simply call the setter
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$this->setBackBuffer('');

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Seeks to beginning of file, updates seek position in this object and
	 * flushes the header.
	 *
	 * @param	$flushHeader	Wether the file's header should be flushed (default: false)
	 * @return	void
	 */
	protected function rewindUpdateSeekPosition (bool $flushHeader = false) {
		// Seek to beginning of file
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: flushHeader=%d - CALLED!', intval($flushHeader)));
		$this->rewind();

		// And update seek position ...
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->updateSeekPosition() ...');
		$this->updateSeekPosition();

		// Flush headers?
		if ($flushHeader) {
			// ... to write it back into the file
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->flushFileHeader() ...');
			$this->flushFileHeader();
		}

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Seeks to old position
	 *
	 * @return	void
	 */
	protected function seekToOldPosition () {
		// Seek to currently ("old") saved position
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$this->seek($this->determineSeekPosition());

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Initializes this file class
	 *
	 * @param	$fileInfoInstance	An instance of a SplFileInfo class
	 * @return	void
	 */
	protected function initFile (SplFileInfo $fileInfoInstance) {
		// Get a file i/o pointer instance
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: fileInfoInstance[%s]=%s - CALLED!', get_class($fileInfoInstance), $fileInfoInstance));
		$pointerInstance = ObjectFactory::createObjectByConfiguredName('file_raw_input_output_class', array($fileInfoInstance));

		// ... and set it here
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Setting pointerInstance=%s ...', $pointerInstance->__toString()));
		$this->setPointerInstance($pointerInstance);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Marks the currently loaded block as empty (with length of the block)
	 *
	 * @param	$length		Length of the block
	 * @return	void
	 * @throws	InvalidArgumentException	If a parameter is invalid
	 */
	protected function markCurrentBlockAsEmpty (int $length) {
		// Validate parameter
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: length=%d - CALLED!', $length));
		if ($length < 1) {
			// Length cannot below one
			throw new InvalidArgumentException(sprintf('length=%d is not valid', $length));
		}

		// Get current seek position
		$currentPosition = $this->determineSeekPosition();

		// Now add it as gap entry
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: currentPosition=%d', $currentPosition));
		$this->addGap(($currentPosition - $length), $currentPosition);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Initializes counter for valid entries, arrays for damaged entries and
	 * an array for gap seek positions. If you call this method on your own,
	 * please re-analyze the file structure. So you are better to call
	 * analyzeFileStructure() instead of this method.
	 *
	 * @return	void
	 */
	public function initCountersGapsArray () {
		// Init counter and seek position to header size
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$this->setCounter(0);
		$this->setSeekPosition($this->getHeaderSize());

		// Init arrays
		$this->gaps = [];
		$this->damagedEntries = [];

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Updates seekPosition attribute from file to avoid to much access on file.
	 *
	 * @return	void
	 */
	public function updateSeekPosition () {
		// Get key (= seek position)
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$seekPosition = $this->determineSeekPosition();

		// And set it here
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: seekPosition=%d', $seekPosition));
		$this->setSeekPosition($seekPosition);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Checks whether the block separator has been found
	 *
	 * @param	$str		String to look in
	 * @return	$isFound	Whether the block separator has been found
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 */
	public static function isBlockSeparatorFound (string $str) {
		// Validate parameter
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: str=%s - CALLED!', $str));
		if (empty($str)) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "str" is empty');
		}

		// Determine it
		$isFound = (strpos($str, chr(self::SEPARATOR_ENTRIES)) !== false);

		// Return result
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: isFound=%d - EXIT!', intval($isFound)));
		return $isFound;
	}

	/**
	 * Writes data at given position
	 *
	 * @param	$seekPosition	Seek position
	 * @param	$data			Data to be written
	 * @param	$flushHeader	Whether to flush the header (default: flush)
	 * @return	void
	 * @throws	OutOfBoundsException	If the position is not seekable
	 * @throws	InvalidArgumentException	If a parameter is invalid
	 */
	public function writeData (int $seekPosition, string $data, bool $flushHeader = true) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: seekPosition=%s,data()=%d,flushHeader=%d - CALLED!', $seekPosition, strlen($data), intval($flushHeader)));
		if ($seekPosition < 0) {
			// Invalid seek position
			throw new OutOfBoundsException(sprintf('seekPosition=%d is not valid', $seekPosition));
		} elseif (empty($data)) {
			// Empty data is invalid, too
			throw new InvalidArgumentException('Parameter "data" is empty');
		}

		// Write data at given position
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Calling this->writeAtPosition(%d,%s) ...', $seekPosition, $data));
		$this->writeAtPosition($seekPosition, $data);

		// Increment counter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->incrementCounter() ...');
		$this->incrementCounter();

		// Update seek position
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->updateSeekPosition() ...');
		$this->updateSeekPosition();

		// Flush the header?
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: flushHeader=%d', intval($flushHeader)));
		if ($flushHeader === true) {
			// Flush header
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->flushFileHeader() ...');
			$this->flushFileHeader();

			// Seek to old position
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->seekToOldPosition() ...');
			$this->seekToOldPosition();
		}

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Writes at given position by seeking to it.
	 *
	 * @param	$seekPosition	Seek position in file
	 * @param	$dataStream		Data to be written
	 * @return	mixed			Number of writes bytes or false on error
	 * @throws	OutOfBoundsException	If the position is not seekable
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 */
	public function writeAtPosition (int $seekPosition, string $dataStream) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: seekPosition=%d,dataStream(%d)=%s - CALLED!', $seekPosition, strlen($dataStream), $dataStream));
		if ($seekPosition < 0) {
			// Invalid seek position
			throw new OutOfBoundsException(sprintf('seekPosition=%d is not valid.', $seekPosition));
		} elseif (empty($dataStream)) {
			// Empty dataStream
			throw new InvalidArgumentException('Parameter "dataStream" is empty');
		}

		// Call pointer's method
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Calling this->pointerInstance->writeAtPosition(%d, %s) ...', $seekPosition, $dataStream));
		$status = $this->getPointerInstance()->writeAtPosition($seekPosition, $dataStream);

		// Return status
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: status[%s]=%d - EXIT!', gettype($status), $status));
		return $status;
	}

	/**
	 * Checks whether the file header is initialized
	 *
	 * @return	$isInitialized	Whether the file header is initialized
	 */
	public function isFileHeaderInitialized () {
		// Default is not initialized
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$isInitialized = false;

		// Is the file initialized?
		if ($this->isFileInitialized()) {
			// Some bytes has been written, so rewind to start of it.
			$this->rewind();

			// Read file header
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->readFileHeader() ...');
			$this->readFileHeader();

			// Get header count
			$headerCount = count($this->getHeader());

			// The above method does already check the header
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: headerCount=%d', $headerCount));
			$isInitialized = ($headerCount > 0);
		}

		// Return result
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: isInitialized=%d - EXIT!', intval($isInitialized)));
		return $isInitialized;
	}

	/**
	 * Checks whether the assigned file has been initialized
	 *
	 * @return	$isInitialized		Whether the file's size is zero
	 * @throws	UnexpectedValueException	If an unexpected value was returned
	 */
	public function isFileInitialized () {
		// Get it from iterator which holds the pointer instance. If false is returned
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$fileSize = $this->size();

		/*
		 * The returned file size should not be false or NULL as this means
		 * that the pointer class does not work correctly.
		 */
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: fileSize[%s]=%d', gettype($fileSize), $fileSize));
		if (!is_int($fileSize)) {
			// Bad file?
			throw new UnexpectedValueException(sprintf('fileSize[]=%s is unexpected', gettype($fileSize)));
		}

		// Is more than 0 returned?
		$isInitialized = ($fileSize > 0);

		// Return result
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: isInitialized=%d - EXIT!', intval($isInitialized)));
		return $isInitialized;
	}

	/**
	 * Creates the assigned file
	 *
	 * @return	void
	 * @throws	BadMethodCallException	If this file's header is already initialized
	 */
	public function createFileHeader () {
		// The file's header should not be initialized here
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		if ($this->isFileHeaderInitialized()) {
			// Bad method call
			//* DEBUG-DIE: */ ApplicationEntryPoint::exitApplication(sprintf('[%s:%d]: this=%s', __METHOD__, __LINE__, print_r($this, TRUE)));
			throw new BadMethodCallException('File header is already initialized but method called');
		}

		// Simple flush file header which will create it.
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->flushFileHeader() ...');
		$this->flushFileHeader();

		// Rewind seek position (to beginning of file) and update/flush file header
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->rewindUpdateSeekPosition() ...');
		$this->rewindUpdateSeekPosition();

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Pre-allocates file (if enabled) with some space for later faster write access.
	 *
	 * @param	$type	Type of the file
	 * @return	void
	 * @throws	InvalidArgumentException	If a parameter is empty
	 */
	public function preAllocateFile (string $type) {
		// Is it enabled?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: type=%s - CALLED!', $type));
		if (empty($type)) {
			// Empty type
			throw new InvalidArgumentException('Parameter "type" is empty');
		} elseif (FrameworkBootstrap::getConfigurationInstance()->getConfigEntry($type . '_pre_allocate_enabled') != 'Y') {
			// Don't continue here.
			self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Not pre-allocating file.'));
			return;
		}

		// Message to user
		self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Pre-allocating file ...');

		// Calculate minimum length for one entry and get file size
		$minimumBlockLength = $this->getIndexInstance()->calculateMinimumBlockLength();
		$fileSize = $this->getFileSize();

		// Calulcate seek position
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: minimumBlockLength=%d,fileSize=%d', $minimumBlockLength, $fileSize));
		$seekPosition = $minimumBlockLength * FrameworkBootstrap::getConfigurationInstance()->getConfigEntry($type . '_pre_allocate_count') + $fileSize ;

		// Now simply write a NUL there. This will pre-allocate the file.
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Calling this->writeAtPosition(%d,NUL) ...', $seekPosition));
		$this->writeAtPosition($seekPosition, chr(0));

		// Is the seek position zero?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: fileSize=%d', $fileSize));
		if ($fileSize == 0) {
			// Mark file as gaps-only
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Calling this->markGapsOnly(%s,%d) ...', $type, $minimumBlockLength));
			$this->markFileGapsOnly($type, $minimumBlockLength);
		} else {
			// Analyze file structure
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->analyzeFileStructure() ...');
			$this->analyzeFileStructure();
		}

		// Rewind seek position
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->rewind() ...');
		$this->rewind();

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Determines seek position
	 *
	 * @return	$seekPosition	Current seek position
	 */
	public function determineSeekPosition () {
		// Call pointer instance
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$seekPosition = $this->getPointerInstance()->determineSeekPosition();

		// Return position
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: seekPosition=%d - EXIT!', $seekPosition));
		return $seekPosition;
	}

	/**
	 * Seek to given offset (default) or other possibilities as fseek() gives.
	 *
	 * @param	$offset		Offset to seek to (or used as "base" for other seeks)
	 * @param	$whence		Added to offset (default: only use offset to seek to)
	 * @return	$status		Status of file seek: 0 = success, -1 = failed
	 * @throws	OutOfBoundsException	If the position is not seekable
	 */
	public function seek (int $offset, int $whence = SEEK_SET) {
		// Validate parameter
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: offset=%d,whence=%d - CALLED!', $offset, $whence));
		if ($offset < 0) {
			// No offset is smaller than zero
			throw new OutOfBoundsException(sprintf('offset=%d is not valid', $offset));
		}

		// Call pointer instance
		$status = $this->getPointerInstance()->seek($offset, $whence);

		// Return status
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: status[%s]=%d - EXIT!', gettype($status), $status));
		return $status;
	}

	/**
	 * Reads given amount of bytes from file.
	 *
	 * @param	$bytes	Amount of bytes to read
	 * @return	$data	Data read from file
	 * @throws	OutOfBoundsException	If the position is not seekable
	 */
	public function read (int $bytes = 0) {
		// Validate parameter
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: bytes=%d - CALLED!', $bytes));
		if ($bytes < 0) {
			// Throw exception
			throw new OutOfBoundsException(sprintf('bytes=%d is not valid', $bytes));
		}

		// Call pointer instance
		$data = $this->getPointerInstance()->read($bytes);

		// Return data
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: data[%s]=%s - EXIT!', gettype($data), $data));
		return $data;
	}

	/**
	 * Rewinds to the beginning of the file
	 *
	 * @return	void
	 */
	public function rewind () {
		// Call pointer instance
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$this->getPointerInstance()->rewind();

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Analyzes entries in index file. This will count all found (and valid)
	 * entries, mark invalid as damaged and count gaps ("fragmentation"). If
	 * only gaps are found, the file is considered as "virgin" (no entries).
	 *
	 * @return	void
	 * @throws	BadMethodCallException	If this method is called but file is not initialized
	 */
	public function analyzeFileStructure () {
		// Make sure the file is initialized
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		if (!$this->isFileInitialized()) {
			// Bad method call
			throw new BadMethodCallException('Method called but file is not initialized.');
		}

		// Init counters and gaps array
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->initCounterGapsArrays() ...');
		$this->initCountersGapsArray();

		// Output message (as this may take some time)
		self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('Analyzing file structure ... (this may take some time)'));

		// First rewind to the begining
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->rewind() ...');
		$this->rewind();

		// Then try to load all entries
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Looping through file');
		while ($this->isValid()) {
			// Get current entry
			$current = $this->getCurrentBlock();

			// Go to next entry
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->readNextBlock() ...');
			$this->readNextBlock();

			/*
			 * If the block is empty, maybe the whole file is? This could mean
			 * that the file has been pre-allocated.
			 */
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: current(%d)[]=%s', strlen($current), gettype($current)));
			if (empty($current)) {
				// Then skip this part
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: current[]=%s is empty - CONTINUE!', gettype($current)));
				continue;
			}

			// Handle current record
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: current(%d)[]=%s', strlen($current), gettype($current)));
		}

		// If the last read block is empty, check gaps
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: current()=%d', strlen($current)));
		if (empty($current)) {
			// Output message
			self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Found a total of %d gaps.', count($this->gaps)));

			// Check gaps, if the whole file is empty.
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->isFileGapsOnly() ...');
			if ($this->isFileGapsOnly()) {
				// Only gaps, so don't continue here.
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: File is gaps-only - EXIT!');
				return;
			}

			/*
			 * The above call has calculated a total size of all gaps. If the
			 * percentage of gaps passes a "soft" limit and last
			 * defragmentation is to far in the past, or if a "hard" limit has
			 * reached, run defragmentation.
			 */
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->isDefragmentationNeeded() ...');
			if ($this->isDefragmentationNeeded()) {
				// Run "defragmentation"
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->doRunDefragmentation() ...');
				$this->doRunDefragmentation();
			}
		}

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Reads next "block" of bytes into $currentBlock field
	 *
	 * @return	void
	 * @throws	BadMethodCallException	If this method was called without prior valid() call
	 */
	public function readNextBlock () {
		// Is there nothing to read?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		if (!$this->isValid()) {
			// Nothing to read
			throw new BadMethodCallException('next() invoked but no valid current block (EOF?)');
		}

		// First calculate minimum block length
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: this->seekPosition=%d', $this->determineSeekPosition()));
		$length = $this->getIndexInstance()->calculateMinimumBlockLength();
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: length=%d', $length));

		// Read possibly back-buffered bytes from previous call of next().
		$data = $this->getBackBuffer();

		/*
		 * Read until a entry/block separator has been found. The next read
		 * "block" may not fit, so this loop will continue until the EOB or EOF
		 * has been reached whatever comes first.
		 */
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: data()=%d', strlen($data)));
		while ((!$this->isEndOfFileReached()) && (empty($data) || !self::isBlockSeparatorFound($data))) {
			// Then read the next possible block
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Calling this->read(%d) ...', $length));
			$block = $this->read($length);

			// Is it all empty?
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: block()=%d,length=%d', strlen($block), $length));
			if (strlen(trim($block, chr(0))) == 0) {
				// Mark this block as empty
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Calling this->markCurrentBlockAsEmpty(%d) ...', strlen($block)));
				$this->markCurrentBlockAsEmpty(strlen($block));

				// Exit look
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Empty block found ... - BREAK!');
				break;
			}

			// At this block then
			$data .= $block;

			// A debug message
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: data(%d)=%s', strlen($data), $data));
		}

		/*
		 * Init back-buffer which is the data that has been found beyond the
		 * separator.
		 */
		$this->initBackBuffer();

		// Is $data empty?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: data(%d)=%s', strlen($data), $data));
		if (empty($data)) {
			// Yes, maybe whole file was ...
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Maybe empty file found - EXIT!');
			return;
		}

		// Separate data
		$dataArray = explode(chr(self::SEPARATOR_ENTRIES), $data);

		// Left part is the actual block, right one the back-buffer data, if found
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: dataArray()=%d', count($dataArray)));
		//* PRINTR-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: dataArray=%s', print_r($dataArray, true)));
		$this->setCurrentBlock($dataArray[0]);

		// Is back buffere data found?
		if (isset($dataArray[1])) {
			// Set back buffer
			$this->setBackBuffer($dataArray[1]);
		}

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Checks wether the current entry is valid (not at the end of the file).
	 * This method will return true if an emptied (nulled) entry has been found.
	 *
	 * @return	$isValid	Whether the next entry is valid
	 * @throws	UnexpectedValueException	If some value is not expected
	 */
	public function isValid () {
		// First calculate minimum block length
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$length = $this->getIndexInstance()->calculateMinimumBlockLength();

		// Short be more than zero!
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: length=%d', $length));
		if ($length < 1) {
			// Throw UVE
			throw new UnexpectedValueException(sprintf('length=%d is not expected', $length));
		}

		// Get current seek position
		$seekPosition = $this->determineSeekPosition();

		// Then try to read it
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: seekPosition=%d', $seekPosition));
		$data = $this->read($length);

		// If some bytes could be read, all is fine
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: data[%s]()=%d', gettype($data), strlen($data)));
		$isValid = ((is_string($data)) && (strlen($data) > 0));

		// Get header size
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: isValid=%d', intval($isValid)));
		$headerSize = $this->getHeaderSize();

		// Is the seek position at or beyond the header?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: seekPosition=%d,headerSize=%d', $seekPosition, $headerSize));
		if ($seekPosition >= $headerSize) {
			// Seek back to old position
			$isValid = ($isValid && $this->seek($seekPosition) === 0);
		} else {
			// Seek directly behind the header
			$isValid = ($isValid && $this->seek($headerSize) === 0);
		}

		// Return result
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: isValid=%d - EXIT!', intval($isValid)));
		return $isValid;
	}

	/**
	 * Reads the file header
	 *
	 * @return	void
	 * @throws	LogicException	If both instances are not set
	 */
	public function readFileHeader () {
		// Is index set or stack?
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		if ($this->getIndexInstance() instanceof Indexable) {
			// Call index instance
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->indexInstance->readIndexHeader() ...');
			$this->getIndexInstance()->readIndexHeader();
		} elseif ($this->getStackInstance() instanceof StackableFile) {
			// Call stacke instance
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Calling this->stackInstance->readStackHeader() ...');
			$this->getStackInstance()->readStackHeader();
		} else {
			// Bad logic?
			throw new LogicException('Wether indexInstance nor stackInstance are set');
		}

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Flushes the file header
	 *
	 * @return	void
	 */
	public function flushFileHeader () {
		// Call block instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$this->getIndexInstance()->flushFileHeader();

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Searches for next suitable gap the given length of data can fit in
	 * including padding bytes.
	 *
	 * @param	$length			Length of raw data
	 * @return	$seekPosition	Found next gap's seek position
	 * @throws	InvalidArgumentException	If the parameter is not valid
	 */
	public function searchNextGap (int $length) {
		// If the file is only gaps, no need to seek
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: length=%d - CALLED!', $length));
		if ($length <= 0) {
			// Throw IAE
			throw new InvalidArgumentException(sprintf('length=%d is not valid', $length));
		} elseif ($this->isFileGapsOnly()) {
			/*
			 * The first empty block is the 2nd one right after the header, so
			 * one byte gap to the header.
			 */
			$seekPosition = ($this->getHeaderSize() + 2);

			// Return position
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: seekPosition=%d - EXIT!', $seekPosition));
			return $seekPosition;
		}

		// @TODO Unfinished
		$this->partialStub('length=' . $length);
	}

}
