<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Stack\File;

// Import framework stuff
use Org\Mxchange\CoreFramework\Filesystem\Stack\FileStacker;
use Org\Mxchange\CoreFramework\Filesystem\File\BaseBinaryFile;
use Org\Mxchange\CoreFramework\Generic\UnsupportedOperationException;
use Org\Mxchange\CoreFramework\Stack\File\StackableFile;
use Org\Mxchange\CoreFramework\Utils\String\StringUtils;

// Import SPL stuff
use \InvalidArgumentException;
use \SplFileInfo;

/**
 * A stack file class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2020 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class StackFile extends BaseBinaryFile implements FileStacker {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates an instance of this File class and prepares it for usage
	 *
	 * @param	$infoInstance	An instance of a SplFileInfo class
	 * @param	$stackInstance	An instance of a StackableFile class
	 * @return	$fileInstance	An instance of this File class
	 */
	public final static function createStackFile (SplFileInfo $infoInstance, StackableFile $stackInstance) {
		// Get a new instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: infoInstance[%s]=%s,stackInstance=%s - CALLED!', get_class($infoInstance), $infoInstance, $stackInstance->__toString()));
		$fileInstance = new StackFile();

		// Set stack instance here for callbacks
		$fileInstance->setStackInstance($stackInstance);

		// Init this abstract file
		$fileInstance->initFile($infoInstance);

		// Return the prepared instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: fileInstance=%s - EXIT!', $fileInstance->__toString()));
		return $fileInstance;
	}

	/**
	 * Writes given value to the file and returns a hash and gap position for it
	 *
	 * @param	$stackName	Group identifier
	 * @param	$value		Value to be added to the stack
	 * @return	$data		Hash and gap position
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 */
	public function writeValueToFile (string $stackName, $value) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: stackName=%s,value[]=%s - CALLED!', $stackName, gettype($value)));
		if (empty($stackName)) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "stackName" is empty');
		} elseif (is_object($value) || is_resource($value)) {
			// Not wanted here
			throw new InvalidArgumentException(sprintf('value[]=%s is not stackable in files', gettype($value)));
		}

		// Encode/convert the value into a "binary format"
		$encoded = StringUtils::encodeData($value);

		// Get a strong hash for the "encoded" data
		$hash = self::hash($encoded);

		// Then write it to the next free gap
		$data = $this->getStackInstance()->writeDataToFreeGap($stackName, $hash, $encoded);

		// Return info
		return $data;
	}

	/**
	 * Writes given raw data to the file and returns a gap position and length
	 *
	 * @param	$stackName	Group identifier
	 * @param	$hash		Hash from encoded value
	 * @param	$encoded	Encoded value to be written to the file
	 * @return	$data		Gap position and length of the raw data
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function writeDataToFreeGap (string $stackName, string $hash, string $encoded) {
		self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('STACK-FILE: stackName=' . $stackName . ',hash=' . $hash . ',encoded()=' . strlen($encoded));
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

}
