<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Filesystem\Pointer\Input;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;
use Org\Mxchange\CoreFramework\Filesystem\BaseFileIo;
use Org\Mxchange\CoreFramework\Filesystem\FileNotFoundException;
use Org\Mxchange\CoreFramework\Filesystem\FileReadProtectedException;
use Org\Mxchange\CoreFramework\Filesystem\Pointer\InputPointer;
use Org\Mxchange\CoreFramework\Generic\NullPointerException;
use Org\Mxchange\CoreFramework\Generic\UnsupportedOperationException;
use Org\Mxchange\CoreFramework\Object\BaseFrameworkSystem;

// Import SPL stuff
use \SplFileInfo;

/**
 * A class for reading files
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2020 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class FrameworkRawFileInputPointer extends BaseFileIo implements InputPointer {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Create a file pointer based on the given file. The file will also
	 * be verified here.
	 *
	 * @param	$infoInstance	An instance of a SplFileInfo class
	 * @throws	FileIoException				If the file is not reachable
	 * @throws	FileReadProtectedException	If the file is not found or cannot be read
	 * @throws	FileNotFoundException		If the file does not exist
	 * @return	void
	 */
	public static final function createFrameworkRawFileInputPointer (SplFileInfo $infoInstance) {
		// Some pre-sanity checks...
		if (!FrameworkBootstrap::isReachableFilePath($infoInstance)) {
			// File cannot be accessed (due to open_basedir restriction)
			throw new FileIoException($infoInstance, self::EXCEPTION_FILE_NOT_REACHABLE);
		} elseif ((!FrameworkBootstrap::isReadableFile($infoInstance)) && (!$infoInstance->isFile())) {
			// File does not exist
			throw new FileNotFoundException($infoInstance, self::EXCEPTION_FILE_NOT_FOUND);
		} elseif ((!FrameworkBootstrap::isReadableFile($infoInstance)) && ($infoInstance->isFile())) {
			// File exists but cannot be read from
			throw new FileReadProtectedException($infoInstance, self::EXCEPTION_FILE_CANNOT_BE_READ);
		}

		// Try to open a handler
		$fileObject = $infoInstance->openFile('rb');
		if ((is_null($fileObject)) || ($fileObject === false)) {
			// Something bad happend
			throw new FileIoException($infoInstance, self::EXCEPTION_FILE_POINTER_INVALID);
		} // END - if

		// Create new instance
		$pointerInstance = new FrameworkRawFileInputPointer();

		// Set file pointer and file name
		$pointerInstance->setFileObject($fileObject);

		// Return the instance
		return $pointerInstance;
	}

	/**
	 * Read data a file pointer
	 *
	 * @return	mixed	The result of fread()
	 * @throws	NullPointerException	If the file pointer instance is not set by setFileObject()
	 * @throws	LogicException	If there is no object being set
	 */
	public function readFromFile () {
		if (is_null($this->getFileObject())) {
			// Pointer not initialized
			throw new NullPointerException($this, self::EXCEPTION_IS_NULL_POINTER);
		} elseif (!is_object($this->getFileObject())) {
			// Pointer is not a valid resource!
			throw new LogicException(sprintf('this->fileObject[]=%s is no object', gettype($this->getFileObject())));
		}

		// Read data from the file pointer and return it
		return $this->read(1024);
	}

	/**
	 * Reads a line, maximum 4096 Bytes from current file pointer
	 *
	 * @return	$data	Read data from file
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function readLine () {
		// Not supported in binary files ...
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Reads given amount of bytes from file.
	 *
	 * @param	$bytes	Amount of bytes to read
	 * @return	$data	Data read from file
	 */
	public function read (int $bytes = 0) {
		// Try to read given characters
		$data = $this->getFileObject()->fread($bytes);

		// Then return it
		return $data;
	}

	/**
	 * Analyzes entries in index file. This will count all found (and valid)
	 * entries, mark invalid as damaged and count gaps ("fragmentation"). If
	 * only gaps are found, the file is considered as "virgin" (no entries).
	 *
	 * @return	void
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function analyzeFileStructure () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Advances to next "block" of bytes
	 *
	 * @return	void
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function next () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Checks wether the current entry is valid (not at the end of the file).
	 * This method will return true if an emptied (nulled) entry has been found.
	 *
	 * @return	$isValid	Whether the next entry is valid
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function valid () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Gets current seek position ("key").
	 *
	 * @return	$key	Current key in iteration
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function key () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

}
